﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2010 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using HeuristicLab.Collections;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Hive;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.PluginInfrastructure;

namespace HeuristicLab.Clients.Hive {

  [Item("Hive Job", "Represents a hive job.")]
  [StorableClass]
  public class HiveJob : NamedItem, IItemTree<HiveJob> {
    protected static object locker = new object();

    public override Image ItemImage {
      get {
        if (job.Id == Guid.Empty) { // not yet uploaded
          return HeuristicLab.Common.Resources.VSImageLibrary.Event;
        } else {
          if (job.State == JobState.Waiting) return HeuristicLab.Common.Resources.VSImageLibrary.ExecutablePrepared;
          else if (job.State == JobState.Calculating) return HeuristicLab.Common.Resources.VSImageLibrary.ExecutableStarted;
          else if (job.State == JobState.Transferring) return HeuristicLab.Common.Resources.VSImageLibrary.ExecutableStarted;
          else if (job.State == JobState.Paused) return HeuristicLab.Common.Resources.VSImageLibrary.ExecutablePaused;
          else if (job.State == JobState.Aborted) return HeuristicLab.Common.Resources.VSImageLibrary.ExecutableStopped;
          else if (job.State == JobState.Failed) return HeuristicLab.Common.Resources.VSImageLibrary.Error;
          else if (job.State == JobState.Finished) return HeuristicLab.Common.Resources.VSImageLibrary.ExecutableStopped;
          else return HeuristicLab.Common.Resources.VSImageLibrary.Event;
        }
      }
    }

    protected Job job;
    public Job Job {
      get { return job; }
      set {
        if (job != value) {
          job = value;
          OnJobChanged();
          OnToStringChanged();
          OnItemImageChanged();
        }
      }
    }

    [Storable]
    protected ItemJob jobItem;
    public ItemJob JobItem {
      get { return jobItem; }
      internal set {
        if (jobItem != null && syncJobsWithOptimizers) {
          this.childHiveJobs.Clear();
        }
        if (jobItem != value) {
          DergisterJobItemsEvents();
          jobItem = value;
          RegisterJobItemEvents();
          OnJobItemChanged();
        }
      }
    }

    [Storable]
    protected ItemList<HiveJob> childHiveJobs;
    public virtual ReadOnlyItemList<HiveJob> ChildHiveJobs {
      get { return childHiveJobs.AsReadOnly(); }
    }

    [Storable]
    protected bool syncJobsWithOptimizers = true;

    #region Constructors and Cloning
    public HiveJob() {
      this.Job = new Job() {
        CoresNeeded = 1,
        MemoryNeeded = 0
      };
      job.State = JobState.Offline;
      this.childHiveJobs = new ItemList<HiveJob>();
      syncJobsWithOptimizers = true;
      RegisterChildHiveJobEvents();
    }

    public HiveJob(ItemJob jobItem, bool autoCreateChildHiveJobs)
      : this() {
      this.syncJobsWithOptimizers = autoCreateChildHiveJobs;
      this.JobItem = jobItem;
      this.syncJobsWithOptimizers = true;
    }

    public HiveJob(Job job, JobData jobData, bool autoCreateChildHiveJobs) {
      this.syncJobsWithOptimizers = autoCreateChildHiveJobs;
      this.Job = job;
      try {
        this.JobItem = PersistenceUtil.Deserialize<ItemJob>(jobData.Data);
      }
      catch {
        this.JobItem = null;
      }
      this.childHiveJobs = new ItemList<HiveJob>();
      this.syncJobsWithOptimizers = true;
      RegisterChildHiveJobEvents();
    }

    protected HiveJob(HiveJob original, Cloner cloner)
      : base(original, cloner) {
      this.Job = cloner.Clone(original.job);
      this.JobItem = cloner.Clone(original.JobItem);
      this.childHiveJobs = cloner.Clone(original.childHiveJobs);
      this.syncJobsWithOptimizers = original.syncJobsWithOptimizers;
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new HiveJob(this, cloner);
    }
    #endregion

    protected virtual void UpdateChildHiveJobs() { }

    protected virtual void RegisterJobItemEvents() {
      if (JobItem != null) {
        JobItem.ComputeInParallelChanged += new EventHandler(JobItem_ComputeInParallelChanged);
        JobItem.ToStringChanged += new EventHandler(JobItem_ToStringChanged);
      }
    }
    protected virtual void DergisterJobItemsEvents() {
      if (JobItem != null) {
        JobItem.ComputeInParallelChanged -= new EventHandler(JobItem_ComputeInParallelChanged);
        JobItem.ToStringChanged -= new EventHandler(JobItem_ToStringChanged);
      }
    }

    protected virtual void RegisterChildHiveJobEvents() {
      this.childHiveJobs.ItemsAdded += new CollectionItemsChangedEventHandler<IndexedItem<HiveJob>>(OnItemsAdded);
      this.childHiveJobs.ItemsRemoved += new CollectionItemsChangedEventHandler<IndexedItem<HiveJob>>(OnItemsRemoved);
      this.childHiveJobs.CollectionReset += new CollectionItemsChangedEventHandler<IndexedItem<HiveJob>>(OnCollectionReset);
    }
    protected virtual void DeregisterChildHiveJobEvents() {
      this.childHiveJobs.ItemsAdded -= new CollectionItemsChangedEventHandler<IndexedItem<HiveJob>>(OnItemsAdded);
      this.childHiveJobs.ItemsRemoved -= new CollectionItemsChangedEventHandler<IndexedItem<HiveJob>>(OnItemsRemoved);
      this.childHiveJobs.CollectionReset -= new CollectionItemsChangedEventHandler<IndexedItem<HiveJob>>(OnCollectionReset);
    }

    protected virtual void JobItem_ToStringChanged(object sender, EventArgs e) {
      this.OnToStringChanged();
    }
    
    protected virtual void JobItem_ComputeInParallelChanged(object sender, EventArgs e) {
      if (JobItem != null && syncJobsWithOptimizers) {
        this.UpdateChildHiveJobs();
      }
    }

    public virtual void AddChildHiveJob(HiveJob hiveJob) {
      this.childHiveJobs.Add(hiveJob);
    }

    public override string ToString() {
      if (jobItem != null) {
        return jobItem.ToString();
      } else {
        return base.ToString();
      }
    }

    public virtual void UpdateFromLightweightJob(LightweightJob lightweightJob) {
      if (lightweightJob != null) {
        job.Id = lightweightJob.Id;
        job.ParentJobId = lightweightJob.ParentJobId;
        job.ExecutionTime = lightweightJob.ExecutionTime;
        job.State = lightweightJob.State;
        job.StateLog = new List<StateLog>(lightweightJob.StateLog);
        job.Command = lightweightJob.Command;
        job.LastJobDataUpdate = lightweightJob.LastJobDataUpdate;
        
        OnJobStateChanged();
        OnToStringChanged();
        OnItemImageChanged();
      }
    }

    /// <summary>
    /// Creates a JobData object containing the Job and the IJob-Object as byte[]
    /// </summary>
    /// <param name="withoutChildOptimizers">
    ///   if true the Child-Optimizers will not be serialized (if the job contains an Experiment)
    /// </param>
    public virtual JobData GetAsJobData(bool withoutChildOptimizers, out List<IPluginDescription> plugins) {
      plugins = null;
      return null;
    }

    #region Events
    public event EventHandler JobChanged;
    private void OnJobChanged() {
      EventHandler handler = JobChanged;
      if (handler != null) handler(this, EventArgs.Empty);
    }

    public event EventHandler JobStateChanged;
    private void OnJobStateChanged() {
      EventHandler handler = JobStateChanged;
      if (handler != null) handler(this, EventArgs.Empty);
    }

    public event EventHandler JobItemChanged;
    private void OnJobItemChanged() {
      JobItem_ComputeInParallelChanged(this, EventArgs.Empty);
      var handler = JobItemChanged;
      if (handler != null) handler(this, EventArgs.Empty);
    }

    public event EventHandler IsFinishedOptimizerDownloadedChanged;
    private void OnIsFinishedOptimizerDownloadedChanged() {
      var handler = IsFinishedOptimizerDownloadedChanged;
      if (handler != null) handler(this, EventArgs.Empty);
    }
    #endregion
    
    /// <summary>
    /// Returns a list of HiveJobs including this and all its child-jobs recursively
    /// </summary>
    public IEnumerable<HiveJob> GetAllHiveJobs() {
      var jobs = new List<HiveJob>();
      jobs.Add(this);
      foreach (HiveJob child in this.ChildHiveJobs) {
        jobs.AddRange(child.GetAllHiveJobs());
      }
      return jobs;
    }

    public HiveJob GetParentByJobId(Guid jobId) {
      if (this.ChildHiveJobs.SingleOrDefault(j => j.job.Id == jobId) != null)
        return this;
      foreach (HiveJob child in this.childHiveJobs) {
        HiveJob result = child.GetParentByJobId(jobId);
        if (result != null)
          return result;
      }
      return null;
    }

    /// <summary>
    /// Searches for an HiveJob object with the correct jobId recursively
    /// </summary>
    public HiveJob GetHiveJobByJobId(Guid jobId) {
      if (this.Job.Id == jobId) {
        return this;
      } else {
        foreach (HiveJob child in this.ChildHiveJobs) {
          HiveJob result = child.GetHiveJobByJobId(jobId);
          if (result != null)
            return result;
        }
      }
      return null;
    }

    public void RemoveByJobId(Guid jobId) {
      IEnumerable<HiveJob> jobs = ChildHiveJobs.Where(j => j.Job.Id == jobId).ToList(); // if Guid.Empty needs to be removed, there could be more than one with this jobId
      foreach (HiveJob j in jobs) {
        this.childHiveJobs.Remove(j);
      }
      foreach (HiveJob child in ChildHiveJobs) {
        child.RemoveByJobId(jobId);
      }
    }

    public IEnumerable<IItemTree<HiveJob>> GetChildItems() {
      return this.childHiveJobs;
    }

    #region INotifyObservableCollectionItemsChanged<IItemTree> Members

    public event CollectionItemsChangedEventHandler<IItemTree<HiveJob>> CollectionReset;
    private void OnCollectionReset(object sender, CollectionItemsChangedEventArgs<IndexedItem<HiveJob>> e) {
      var handler = CollectionReset;
      if (handler != null) handler(this, ToCollectionItemsChangedEventArgs(e));
    }

    public event CollectionItemsChangedEventHandler<IItemTree<HiveJob>> ItemsAdded;
    private void OnItemsAdded(object sender, CollectionItemsChangedEventArgs<IndexedItem<HiveJob>> e) {
      var handler = ItemsAdded;
      if (handler != null) handler(this, ToCollectionItemsChangedEventArgs(e));
    }

    public event CollectionItemsChangedEventHandler<IItemTree<HiveJob>> ItemsRemoved;
    private void OnItemsRemoved(object sender, CollectionItemsChangedEventArgs<IndexedItem<HiveJob>> e) {
      var handler = ItemsRemoved;
      if (handler != null) handler(this, ToCollectionItemsChangedEventArgs(e));
    }

    private static CollectionItemsChangedEventArgs<IItemTree<HiveJob>> ToCollectionItemsChangedEventArgs(CollectionItemsChangedEventArgs<IndexedItem<HiveJob>> e) {
      return new CollectionItemsChangedEventArgs<IItemTree<HiveJob>>(e.Items.Select(x => x.Value), e.OldItems == null ? null : e.OldItems.Select(x => x.Value));
    }
    #endregion

    public void Pause() {
      ServiceLocator.Instance.CallHiveService(s => s.PauseJob(this.job.Id));
    }

    public void Stop() {
      ServiceLocator.Instance.CallHiveService(s => s.StopJob(this.job.Id));
    }

    public void Restart() {
      ServiceLocator.Instance.CallHiveService(service => {
        JobData jobData = new JobData();
        jobData.JobId = this.job.Id;
        jobData.Data = PersistenceUtil.Serialize(this.jobItem);
        service.UpdateJobData(this.Job, jobData);
        service.RestartJob(this.job.Id);
        Job job = service.GetJob(this.job.Id);
        this.job.LastJobDataUpdate = job.LastJobDataUpdate;
      });
    }
    
    public ICollection<IItemTreeNodeAction<HiveJob>> Actions {
      get {
        return new List<IItemTreeNodeAction<HiveJob>>();
      }
    }
  }

  [Item("Hive Job", "Represents a hive job.")]
  [StorableClass]
  public class HiveJob<T> : HiveJob where T : ItemJob {

    public new T JobItem {
      get { return (T)base.JobItem; }
      internal set { base.JobItem = value; }
    }

    #region Constructors and Cloning
    public HiveJob() : base() { }
    protected HiveJob(HiveJob original, Cloner cloner)
      : base(original, cloner) {
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new HiveJob<T>(this, cloner);
    }
    #endregion
  }
}